using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;
using System.IO;

namespace ShoreSide.Core
{
    //this class prepares a hex file for the Tiny Bootloader running on the main PIC
    public class BootLoader
    {
        public static bool ReadyToLoad = false;
        public static string FileName = null;
        const int MaxFlashBytes = 0x4000 * 2;
        const int SizeOfBLD = 200;
        const int StartBldZone = MaxFlashBytes - SizeOfBLD;

        static byte[] Flash = new Byte[MaxFlashBytes];
        static bool[] FlashDirty = new bool[MaxFlashBytes];

        static byte[] EEPROM = new byte[0x4000];
        static bool[] EEPROMDirty = new bool[0x4000];

        static byte[] Config = new byte[0x1000];
        static bool[] ConfigDirty = new bool[0x1000];

        public static void Clear()
        {
            Flash = new Byte[MaxFlashBytes];
            FlashDirty = new bool[MaxFlashBytes];

            EEPROM = new byte[0x4000];
            EEPROMDirty = new bool[0x4000];

            Config = new byte[0x1000];
            ConfigDirty = new bool[0x1000];
        }

        public static void PrepareHex()
        {
            for (int i = 0; i < 8; i++)
            {
                Flash[StartBldZone + i] = Flash[i];
                FlashDirty[StartBldZone + i] = true;
                FlashDirty[i] = false;
            }

            for (int i = StartBldZone + 8; i < StartBldZone + SizeOfBLD; i++)
                if (FlashDirty[i])
                    throw new Exception("Program overrites Bootloader!");

            int address = (StartBldZone + 8) / 2;
            int i1 = (0xEF00 | (address % 256));  // put a goto Bootloader in the first instructions
            int i2 = (0xF000 | (address / 256));
            Flash[0] = (byte)(i1 % 256);
            Flash[1] = (byte)(i1 / 256);
            Flash[2] = (byte)(i2 % 256);
            Flash[3] = (byte)(i2 / 256);

            for (int i = 0; i < 4; i++)
                FlashDirty[i] = true;
        }

        public static bool Connect(NetworkStream stream)
        {
            byte[] b = new byte[2];
            Timer timer = new Timer();
            
            stream.Flush();
            while (stream.DataAvailable)
                stream.ReadByte();

            stream.WriteByte(0xC1);
            timer.Start();
            stream.Flush();

            int bytesRead = 0;
            timer.Stop();
            while (bytesRead < 2 && timer.Duration < 3000)
            {
                if (stream.DataAvailable)
                    bytesRead = stream.Read(b, bytesRead, 2 - bytesRead);
                timer.Stop();
            }

            if (bytesRead == 2)
                return true;
            else
                return false;


        }

        public static bool WriteHex(NetworkStream stream)
        {
            int address = 0;
            int QBlocks = 0;
            int TotalQBlocks = 0;

            for (int i = 0; i < MaxFlashBytes / 64; i++)
                if (!Next64AreEmpty(i * 64))
                    TotalQBlocks++;

            while (address < MaxFlashBytes)
            {
                while (address < MaxFlashBytes && Next64AreEmpty(address))
                    address += 64;

                if (address < MaxFlashBytes)
                {
                    int count = 64;
                    byte crc = 0;
                    byte[] buffer = new Byte[count + 5];

                    buffer[0] = (byte)(address / (int)0x10000);
                    buffer[1] = (byte)(address / (int)256);
                    buffer[2] = (byte)(address % 256);
                    buffer[3] = (byte)count;

                    Array.Copy(Flash, address, buffer, 4, count);

                    for (int i = 0; i < count; i++)
                        crc = (byte)(crc + buffer[i + 4]);

                    crc = (byte)(256 - (crc + buffer[0] + buffer[1] + buffer[2] + buffer[3]));
                    buffer[buffer.Length - 1] = crc;

                    stream.Write(buffer, 0, buffer.Length);

                    byte[] b = new byte[1];

                    System.Threading.Thread.Sleep(80);

                    if (stream.DataAvailable)
                        stream.Read(b, 0, 1);
                    else
                        b[0] = (byte)'K';

                    if (b[0] == 'K')
                    {
                        address += 64;
                        QBlocks++;
                        DBG(string.Format("FIRMWARE: {0} of {1} QBlocks written", QBlocks, TotalQBlocks));
                    }
                    else if (b[0] == 'N')
                    {
                        DBG(string.Format("FIRMWARE: CRC error retransmitting"));
                    }
                    else
                    {
                        DBG(string.Format("FIRMWARE: Unexpected CHAR! '{0}' Quitting", b[0]));
                        return false;
                    }
                }
            }
            return true;
        }

        private static bool Next64AreEmpty(int address)
        {
            for (int i = address; i < address + 64; i++)
                if (FlashDirty[i])
                    return false;

            return true;
        }

        static BootLoader()
        {
        }

        private static int GetByteAddressType(int address, ref int relByteAddress)
        {
            if (address >= 0 && address < 0x200000)
            {
                relByteAddress = 0;
                return 1;
            }
            else if (address >= 0xF00000 && address < 0xF10000)
            {
                relByteAddress = address - 0xF00000;
                return 2;
            }
            else if (address >= 0x300000 && address < 0x400000)
            {
                relByteAddress = address - 0x300000;
                return 3;
            }
            else
            {
                throw new Exception("Unknown Address Type");
            }
        }
        private static byte ParseHexDigit(char c)
        {
            if (c >= '0' && c <= '9')
            {
                return (byte)(c - '0');
            }
            if (c >= 'a' && c <= 'f')
            {
                return (byte)(c - 'a' + 10);
            }
            if (c >= 'A' && c <= 'F')
            {
                return (byte)(c - 'A' + 10);
            }
            throw new ArgumentException("Invalid hex character");
        }

        private static byte[] ParseHex(string hex)
        {
            byte[] result = new byte[hex.Length / 2];
            int hexIndex = 0;

            for (int i = 0; i < result.Length; i++)
                result[i] = (byte)(ParseHexDigit(hex[hexIndex++]) * 16 + ParseHexDigit(hex[hexIndex++]));

            return result;
        }


        public static bool ParseHexFile(string fileName)
        {
            if (!File.Exists(fileName))
                return false;

            FileName = fileName;

            StreamReader file = File.OpenText(FileName);

            int addressPrefix = 0;
            int hexByteCount = 0;
            string line = null;

            try
            {
                while ((line = file.ReadLine()) != null)
                {
                    if (!line.StartsWith(";"))	//skip if the line is a comment
                    {
                        int bytes = ParseHex(line.Substring(1, 2))[0];
                        byte[] adrBytes = ParseHex(line.Substring(3, 4));
                        int lineAddress = adrBytes[0] * 256 + adrBytes[1];
                        int recordType = ParseHex(line.Substring(7, 2))[0];

                        byte[] record = ParseHex(line.Substring(9, bytes * 2));


                        switch (recordType)
                        {
                            case 0:
                                {
                                    int blockAddress = addressPrefix + lineAddress;
                                    hexByteCount += record.Length;

                                    int relativeAddress = 0;

                                    switch (GetByteAddressType(blockAddress, ref relativeAddress))
                                    {
                                        case 1:
                                            if (blockAddress < MaxFlashBytes)
                                            {
                                                Array.Copy(record, 0, Flash, blockAddress, record.Length);
                                                for (int i = blockAddress; i < blockAddress + record.Length; i++)
                                                    FlashDirty[i] = true;
                                            }
                                            break;
                                        case 2:
                                            Array.Copy(record, 0, EEPROM, relativeAddress, record.Length);
                                            for (int i = relativeAddress; i < relativeAddress + record.Length; i++)
                                                EEPROMDirty[i] = true;
                                            break;
                                        case 3:
                                            Array.Copy(record, 0, Config, relativeAddress, record.Length);
                                            for (int i = relativeAddress; i < relativeAddress + record.Length; i++)
                                                ConfigDirty[i] = true;
                                            break;
                                    }

                                    break;
                                }
                            case 4:	//Extended block
                                addressPrefix = (record[0] * 256 + record[1]) * 0x10000;
                                break;
                            case 1:	//ignoring
                                break;

                        }

                        DBG(string.Format("bytes = {0} adr = {1} rectype = {2}", bytes, lineAddress, recordType));
                    }
                }
            }
            catch (Exception)
            {
                DBG("Bad Firmware HEX file");
            }

            try
            {
                file.Close();
            }
            catch (Exception)
            { }
            return true;
        }

        private static void DBG(string s)
        {
            Controller.Trace(s);
        }
    }
}
