using System;
using System.Collections.Generic;
using System.Text;
using Sharp3D.Math.Core;
using Sharp3D.Math.Geometry3D;

namespace ShoreSide.Core
{
    public class Leg
    {
        public int LegIndex = 0;
        public int ShoulderYawIndex = 0;
        public int ShoulderPitchIndex = 0;
        public int ElbowIndex = 0;

        public const double Seg1Length = 30, Seg2Length = 70, Seg3Length = 110;

        private double x = 0, y = 0, z = 0;
        private double j1 = 0, j2 = 0, j3 = 0;

        private bool positionValid = false; //Calc Angles sets this 
                                // to indicate whether the point can
                                // or cannot be reached


        public void Apply()
        {
            Servo.ServosByIndex[ShoulderYawIndex].Angle = ShoulderYawAngle;
            Servo.ServosByIndex[ShoulderPitchIndex].Angle = ShoulderPitchAngle;
            Servo.ServosByIndex[ElbowIndex].Angle = ElbowAngle;
        }

        public void ReadPositions()
        {
            ReadPositions(this);
        }

        public void ReadCommandedPositions()
        {
            ReadCommandedPositions(this);
        }

        private static void ReadPositions(Leg leg)
        {
            leg.SetAngles(Servo.ServosByIndex[leg.ShoulderYawIndex].AngleReading, Servo.ServosByIndex[leg.ShoulderPitchIndex].AngleReading, Servo.ServosByIndex[leg.ElbowIndex].AngleReading);
        }

        private static void ReadCommandedPositions(Leg leg)
        {
            leg.SetAngles(Servo.ServosByIndex[leg.ShoulderYawIndex].Angle, Servo.ServosByIndex[leg.ShoulderPitchIndex].Angle, Servo.ServosByIndex[leg.ElbowIndex].Angle);
        }

        public static Leg FromReadings(int LegIndex)
        {
            Leg l = new Leg(LegIndex);
            ReadPositions(l);
            return l;
        }

        public static Leg FromCommandedPositions(int LegIndex)
        {
            Leg l = new Leg(LegIndex);
            ReadCommandedPositions(l);
            return l;
        }

        public Leg(int LegIndex)
        {
            this.LegIndex = LegIndex;
            ShoulderYawIndex = Body.ServoMap(LegIndex, Body.Joints.ShoulderYaw);
            ShoulderPitchIndex = Body.ServoMap(LegIndex, Body.Joints.ShoulderPitch);
            ElbowIndex = Body.ServoMap(LegIndex, Body.Joints.Elbow);
        }

        public void CalculateCoordinates()
        {
            Vector4D Position = Matrix4D.Transform(
                TransformationD.RotateY(j1 * Math.PI / 180) *
                TransformationD.Translation(Seg1Length, 0, 0) *
                TransformationD.RotateZ(-j2 * Math.PI / 180) *
                TransformationD.Translation(Seg2Length, 0, 0) *
                TransformationD.RotateZ(j3 * Math.PI / 180 + Math.PI / 2) *
                TransformationD.Translation(Seg3Length, 0, 0)
                , new Vector4D(0, 0, 0, 1));

            x = Position.X;
            y = Position.Y;
            z = Position.Z;
        }

        public Vector3F ShoulderYawPosition
        {
            get
            {
                return new Vector3F(0f, 0f, 0f);
            }
        }

        public Vector3F ShoulderPitchPosition
        {
            get
            {
                Vector4F Position = Matrix4F.Transform(
                    TransformationF.RotateY((float)(z * Math.PI / 180)) *
                    TransformationF.Translation((float)Seg1Length, 0, 0)
                    , new Vector4F(0, 0, 0, 1));

                return new Vector3F((float)Position.X, (float)Position.Y, (float)Position.Z);
            }
        }

        public Vector3F ElbowPosition
        {
            get
            {
                Vector4F Position = Matrix4F.Transform(
                    TransformationF.RotateY((float)(j1 * Math.PI / 180)) *
                    TransformationF.Translation((float)Seg1Length, 0, 0) *
                    TransformationF.RotateZ((float)(-j2 * Math.PI / 180)) *
                    TransformationF.Translation((float)Seg2Length, 0, 0)
                    , new Vector4F(0, 0, 0, 1));
    
                return new Vector3F(Position.X, Position.Y, Position.Z);
            }
        }

        public void CalculateAngles()
        {
            double x, y, z;
            double Joint1Angle, Joint2Angle, Joint3Angle;
            double shoulderX = 0, shoulderY = 0;
            x = this.x;
            y = this.y;
            z = this.z;


            //Find the angle for the first joint
            Joint1Angle = -Math.Atan2(z, x);

            //Rotate the point in the x,z plane around orgin
            x = (x * Math.Cos(Joint1Angle) - z * Math.Sin(Joint1Angle));
            z = (z * Math.Cos(Joint1Angle) + x * Math.Sin(Joint1Angle));

            //Set the shoulder position along the x axis
            shoulderX = Seg1Length;

            //Start to find the intersection of the two circles
            //according to the algo found at http://mathforum.org/library/drmath/view/51836.html
            double dx = x - shoulderX;
            double dy = y - shoulderY;
            double distance = Math.Sqrt(Math.Pow(dx, 2) + Math.Pow(dy, 2)); //Distance between centers

            //Is the distance further than our reach?
            if (distance > (Seg2Length + Seg3Length))
            {
                positionValid = false;
                return;
            }

            positionValid = true;

            //Length of the line connecting the two intersecting points
            double dCenterLine = (Math.Pow(distance, 2) + Math.Pow(Seg2Length, 2) - Math.Pow(Seg3Length, 2)) / (2 * distance);

            double KneeX = shoulderX + dx * dCenterLine / distance + (dy / distance) * Math.Sqrt(Math.Pow(Seg2Length, 2) - Math.Pow(dCenterLine, 2));
            double KneeY = shoulderY + dy * dCenterLine / distance - (dx / distance) * Math.Sqrt(Math.Pow(Seg2Length, 2) - Math.Pow(dCenterLine, 2));
            if (KneeX == Double.NaN || KneeY == Double.NaN)
                return;

            Joint2Angle = Math.Atan2((KneeY - shoulderY), (KneeX - shoulderX));
            Joint3Angle = Math.Atan2((y - KneeY), (x - KneeX));

            j1 = ((180 / Math.PI * Joint1Angle)) % 360;
            j2 = (-(180 / Math.PI * Joint2Angle)) % 360; //Because of the alignment, J2 needs to be inverted
            j3 = ((180 / Math.PI * (Joint3Angle - Joint2Angle)) - 90) % 360; //Because J3 is attached to and rotated from J2

            return;
        }

        public double ShoulderYawAngle
        {
            get
            {
                return j1;
            }
            set
            {
                j1 = value;
                CalculateCoordinates();
            }
        }

        public double ShoulderPitchAngle
        {
            get
            {
                return j2;
            }
            set
            {
                j2 = value;
                CalculateCoordinates();
            }
        }

        public double ElbowAngle
        {
            get
            {
                return j3;
            }
            set
            {
                j3 = value;
                CalculateCoordinates();
            }
        }

        public void SetFootPositionLocal(double X, double Y, double Z)
        {
            x = X;
            y = Y;
            z = Z;
            CalculateAngles();
        }

        public void SetAngles(double ShoulderYaw, double ShoulderPitch, double Elbow)
        {
            j1 = ShoulderYaw;
            j2 = ShoulderPitch;
            j3 = Elbow;
            CalculateCoordinates();
        }

        public double X
        {
            get
            {
                return x;
            }
            set
            {
                x = value;
                CalculateAngles();
            }
        }

        public double Y
        {
            get
            {
                return y;
            }
            set
            {
                y = value;
                CalculateAngles();
            }
        }

        public double Z
        {
            get
            {
                return z;
            }
            set
            {
                z = value;
                CalculateAngles();
            }
        }
        
        public bool CanReach
        {
            get
            {
                return positionValid;
            }
        }

        public Vector3F LegLocalToWorld(Vector3F pos)
        {
            Vector4F point = new Vector4F(pos.X, -pos.Y, -pos.Z, 1f);
            Vector4F newPoint = Matrix4F.Transform(Matrix4F.Multiply(TransformationF.RotateY((float)this.LegIndex * (float)Math.PI / 3), TransformationF.Translation(135f, 0, 0)), point);

            return new Vector3F(newPoint.X, newPoint.Y, newPoint.Z);
        }

        public Vector3F WorldFootPosition()
        {
            Vector4F point = new Vector4F((float)this.X, (float)-this.Y, (float)-this.Z, 1f);
            Vector4F newPoint = Matrix4F.Transform(Matrix4F.Multiply(TransformationF.RotateY((float)this.LegIndex * (float)Math.PI / 3), TransformationF.Translation(135f, 0, 0)), point);

            return new Vector3F(newPoint.X, newPoint.Y, newPoint.Z);
        }

        public Vector3F DisplacementVector()
        {
            float DisplacementX, DisplacementY, DisplacementZ;

            DisplacementX = 100;
            DisplacementY = 0;
            DisplacementZ = 0;

            Vector4F point = new Vector4F(DisplacementX, DisplacementY, DisplacementZ, 1f);
            Vector4F newPoint = Matrix4F.Transform(Matrix4F.Multiply(TransformationF.RotateY((float)this.LegIndex * (float)Math.PI / 3), TransformationF.Translation(135f, 0, 0)), point);

            Vector3F NonDisplacedFoot = new Vector3F(newPoint.X, newPoint.Y, newPoint.Z);
            return WorldFootPosition() - NonDisplacedFoot;
        }
    }
}
