using System;
using System.Collections.Generic;
using System.Text;

namespace ShoreSide.Core
{
    public class PCTLinker
    {
        public static PCTServo[] Servos = null;
        public enum Legs
        {
            LeftFore = 0,
            LeftMid = 1,
            LeftRear = 2,
            RightRear = 3,
            RightMid = 4,
            RightFore = 5
        }

        public enum Joints
        {
            ShoulderYaw = 0,
            ShoulderPitch = 1,
            Elbow = 2
        }

        static PCTLinker()
        {
            Servos = new PCTServo[18];
            for (int i = 0; i < 18; i++)
                Servos[i] = new PCTServo(Servo.ServosByIndex[i]);
        }

        public static void ZeroServos()
        {
            foreach (PCTServo s in Servos)
                s.Reference = 0;
        }
    }

    public class PCTLinkage
    {
        public int Sign = 0;
        public PCTController Controller;

        public PCTLinkage(int sign, PCTController controller)
        {
            Sign = sign;
            Controller = controller;
        }
    }

    public class PCTControllers
    {
        public static HeightController Height = new HeightController();
        public static ForwardController Forward = new ForwardController();
        public static RightController Right = new RightController();
        public static HeadingController Heading = new HeadingController();
        public static PitchController Pitch = new PitchController();
        public static RollController Roll = new RollController();
        public static RadiusController Radius = new RadiusController();
        public static StepHeightController StepHeight = new StepHeightController();

        public static List<LandReflex> LandReflexes = new List<LandReflex>();

        static PCTControllers()
        {
            for (int i = 0; i < 6; i++)
            {
                LandReflex lr = new LandReflex(i);
                lr.Enabled = false;
                LandReflexes.Add(lr);
            }
        }

        public static void Cycle()
        {
            //Reset all the servo reference to zero
            //This needs to be extended once we have more levels and they 
            //are zeroed here because of the crude nature of my accumulators
            foreach (PCTServo s in PCTLinker.Servos)
                s.Reference = 0;

            foreach (PCTController controller in PCTController.AllControllers)
                controller.Cycle();
        }
    }

    public abstract class PCTController
    {
        public const float DEFAULT_P_GAIN = 0.03f;
        public const float DEFAULT_D_GAIN = 0.01f;
        public double P_Gain = DEFAULT_P_GAIN;
        public double D_Gain = DEFAULT_D_GAIN;
        public double OutputMinimum = float.MinValue;
        public double OutputMaximum = float.MaxValue;
        public double Reference = 0;
        private double LastPerception = 0;

        public double Error = 0;
        public double Output = 0;

        public bool Enabled = false;

        public string Name = "Unset";
        public double MinValue = double.MinValue;
        public double MaxValue = double.MaxValue;

        public List<PCTLinkage> Linkages = new List<PCTLinkage>();

        public static List<PCTController> AllControllers = new List<PCTController>();

        public PCTController()
        {
            AllControllers.Add(this);
        }

        public void AddLink(int Sign, PCTLinker.Legs Leg, Body.Joints Joint)
        {
            Linkages.Add(new PCTLinkage(Sign, PCTLinker.Servos[Body.ServoMap((int)Leg, Joint)]));
        }

        public virtual double Perception()
        {
            //return the perception
            return 0;
        }

        public double GetReference()
        {
            //For histograms
            return Reference;
        }

        public void Cycle()
        {
            double perception = Perception();

            double p = 0, d = 0;

            if (!Enabled)
                return;

            //Find error
            Error = Reference - perception;

            //find proportional and differential terms
            p = P_Gain * Error;
            d = D_Gain * (perception - LastPerception);

            LastPerception = perception;

            //compute output
            Output = p - d;
            Output = Math.Max(Output, OutputMinimum);
            Output = Math.Min(Output, OutputMaximum);


            OutputAdjust();
            DoLinkage();
        }

        protected virtual void OutputAdjust()
        {
        }

        protected virtual void DoLinkage()
        {
            foreach (PCTLinkage link in Linkages)
                link.Controller.Reference += Output * link.Sign;
        }
    }
    /*
        Current Linkages

        		    LeftFore    LeftMid     LeftRear    RightRear   RightMid    RightFore

        Height		0--		    0--		    0--		    0--		    0--		    0--
        Forward		+0-		    +00		    +0+		    -0+		    -00		    -0-
        Right		+0+		    00+		    -0+		    -0-		    00-		    +0-

        Heading		+00		    +00		    +00		    +00		    +00		    +00
        Pitch		0--		    000		    0++		    0++		    000		    0--
        Roll		0++		    0++		    0++		    0--		    0--		    0--
        
        Radius		00-		    00-		    00-		    00-		    00-		    00-

        Landing		0--		    0--		    0--		    0--		    0--		    0--

    */

    public class HeightController : PCTController
    {
        public HeightController()
            : base()
        {
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftMid, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.LeftMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);

            this.Name = "Height";
            this.MinValue = -30;
            this.MaxValue = 200;
            this.Reference = 75;
        }

        public override double Perception()
        {
            return Perceptions.Height;
        }
    }

    public class ForwardController : PCTController
    {
        public ForwardController()
            : base()
        {
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderYaw);
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftMid, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);

            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.ShoulderYaw);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderYaw);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);

            this.Name = "Forward";
            this.MinValue = -40;
            this.MaxValue = 40;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.Forwards;
        }
    }

    public class HeadingController : PCTController
    {
        public HeadingController()
            : base()
        {
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.LeftMid, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.RightMid, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderYaw);

            this.Name = "Heading";
            this.MinValue = -40;
            this.MaxValue = 40;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.Heading;
        }
    }

    public class RightController : PCTController
    {
        public RightController()
            : base()
        {
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderYaw);
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftMid, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderYaw);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderYaw);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderYaw);

            this.Name = "Right";
            this.MinValue = -40;
            this.MaxValue = 40;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.Right;
        }
    }

    public class PitchController : PCTController
    {
        public PitchController()
            : base()
        {
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderPitch);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderPitch);
            AddLink(+1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderPitch);

            this.Name = "Pitch";
            this.MinValue = -10;
            this.MaxValue = 10;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.Pitch;
        }
    }

    public class RollController : PCTController
    {
        public RollController()
            : base()
        {
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftFore, Body.Joints.ShoulderPitch);
            AddLink(+1, PCTLinker.Legs.LeftMid, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftMid, Body.Joints.ShoulderPitch);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);
            AddLink(+1, PCTLinker.Legs.LeftRear, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.ShoulderPitch);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.ShoulderPitch);

            this.Name = "Roll";
            this.MinValue = -10;
            this.MaxValue = 10;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.Roll;
        }
    }

    public class LandReflex : PCTController
    {
        public int LegIndex = 0;
        public double LandingSpeed = .5;
        public double ContactThreshold = 128;

        public LandReflex(int legIndex)
            : base()
        {
            this.Name = "Landing[" + legIndex.ToString() + "]";
            this.MinValue = 0;
            this.MaxValue = 1;
            this.Reference = 0;

            this.LegIndex = legIndex;

            AddLink(-1, (PCTLinker.Legs)legIndex, Body.Joints.ShoulderPitch);
            AddLink(-1, (PCTLinker.Legs)legIndex, Body.Joints.Elbow);

        }

        public double RawValue
        {
            get
            {
                return Perceptions.FootSensors[this.LegIndex];
            }
        }

        public override double Perception()
        {
            int FootValue = Perceptions.FootSensors[this.LegIndex];

            if (FootValue > ContactThreshold)
                return 1;
            else
                return 0;
        }

        protected override void OutputAdjust()
        {
            if (Perception() == 0)
            {
                Output = LandingSpeed;
            }
            else
                Output = 0;
        }
    }

    public class PCTServo : PCTController, IDisposable
    {
        private Servo servo = null;
        private Controller.CycleHandler CycleDelegate = null;
        public PCTServo(Servo s)
            : base()
        {
            servo = s;
            CycleDelegate = new Controller.CycleHandler(cycle);
            Controller.Cycle += CycleDelegate;
        }

        public void cycle()
        {
            //update servo position
            servo.Angle += Reference;
            servo.Angle = Math.Min(servo.Angle, 90);
            servo.Angle = Math.Max(servo.Angle, -90);
        }

        #region IDisposable Members

        public void Dispose()
        {
            Delegate.Remove(Controller.Cycle, CycleDelegate);
        }

        #endregion
    }

    public class RadiusController : PCTController
    {
        public RadiusController()
            : base()
        {
            AddLink(-1, PCTLinker.Legs.LeftFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.LeftRear, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightFore, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightMid, Body.Joints.Elbow);
            AddLink(-1, PCTLinker.Legs.RightRear, Body.Joints.Elbow);

            this.Name = "Radius";
            this.MinValue = 50;
            this.MaxValue = 200;
            this.Reference = 100;
        }

        public override double Perception()
        {
            return Perceptions.Radius;
        }

    }

    public class StepHeightController : PCTController
    {
        public StepHeightController()
            : base()
        {
            foreach (int LegIndex in Perceptions.LegGroup0)
            {
                AddLink(+1, (PCTLinker.Legs)LegIndex, Body.Joints.ShoulderPitch);
                AddLink(+1, (PCTLinker.Legs)LegIndex, Body.Joints.Elbow);
            }

            foreach (int LegIndex in Perceptions.LegGroup1)
            {
                AddLink(-1, (PCTLinker.Legs)LegIndex, Body.Joints.ShoulderPitch);
                AddLink(-1, (PCTLinker.Legs)LegIndex, Body.Joints.Elbow);
            }

            this.Name = "StepHeight";
            this.MinValue = -50;
            this.MaxValue = 50;
            this.P_Gain = 0.002f;
            this.Reference = 0;
        }

        public override double Perception()
        {
            return Perceptions.StepHeight;
        }
    }
}
