using System;
using System.Collections.Generic;
using System.Text;
using Sharp3D.Math.Core;
using Sharp3D.Math.Geometry3D;

namespace ShoreSide.Core
{
    class Perceptions
    {
        public static Plane GroundPlane;
        public static Vector3F Centroid;
        public static float Height = 0;
        public static float Forwards = 0;
        public static float Right = 0;
        public static float Heading = 0;
        public static float Pitch = 0;
        public static float Roll = 0;
        public static float Radius = 0;
        public static float StepHeight = 0;

        public static Body Posture = null;
        public static Vector3F[] Feet = new Vector3F[6];

        public static int[] FootSensors = new int[6];

        public static int[] LegGroup0 = new int[] { 0, 2, 4 };
        public static int[] LegGroup1 = new int[] { 1, 3, 5 };

        static Perceptions()
        {
            //Sign up for an update every cycle
            Controller.Cycle += new Controller.CycleHandler(Cycle);
        }

        public static void Cycle()
        {
            Posture = Body.FromReadings();

            for (int i = 0; i < 6; i++)
                Feet[i] = Posture.Legs[i].WorldFootPosition();

            FindGroundPlane();
            FindCentroid();
            FindHeight();
            FindPitchAndRoll();
            FindForwardsAndRight();
            FindHeading();
            FindRadius();
            FindStepHeight();

            DoFootMap();
        }

        private static void FindGroundPlane()
        {
            //Average two alternate triangles to determine ground plane
            Plane p1 = new Plane(Feet[0], Feet[2], Feet[4]);
            Plane p2 = new Plane(Feet[1], Feet[3], Feet[5]);

            Vector3F Normal = new Vector3F((p1.Normal.X + p2.Normal.X) / 2, (p1.Normal.Y + p2.Normal.Y) / 2, (p1.Normal.Z + p2.Normal.Z) / 2);
            float PlaneConstant = (p1.Constant + p2.Constant) / 2;

            GroundPlane = new Plane(Normal, PlaneConstant);
        }

        private static void FindCentroid()
        {
            //Average the displacement vectors to find the centroid
            float x, y, z;
            Vector3F displacement;

            x = y = z = 0;
            for (int i = 0; i < 6; i++)
            {
                displacement = Posture.Legs[i].DisplacementVector();
                x += displacement.X;
                y += displacement.Y;
                z += displacement.Z;
            }

            x /= 6;
            y /= 6;
            z /= 6;

            Centroid = new Vector3F(x, y, z);
        }

        private static void FindHeight()
        {
            //Find the intersection between a line extended from the deck through the ground plane
            IntersectionPair intersection = IntersectionMethods.Intersects(new Line(new Vector3F(0, 1000, 0), new Vector3F(0, -1000, 0)), GroundPlane);
            if (intersection.IntersectionOccurred)
                Height = -intersection.IntersectionPoint.Y;
            else
            {
                //Bad
                Height = 0;
                System.Diagnostics.Trace.WriteLine("Height Perception invalid");
            }
        }

        private static void FindPitchAndRoll()
        {
            //Pitch and Roll are relative to the world X and Z axes
            //Any rotation has to happen before ground plane calculation

            Pitch = (float)(GroundPlane.Normal.X * 180 / Math.PI);
            Roll = (float)(GroundPlane.Normal.Z * 180 / Math.PI);
        }

        private static void FindForwardsAndRight()
        {
            //Again these two are relative to the coordinate system
            //and rotations must be done before calculating the centroid

            Forwards = Centroid.X;
            Right = Centroid.Z;
        }

        private static void FindHeading()
        {
            //Average the yaw angles to determine heading
            float heading = 0;

            for (int i = 0; i < 6; i++)
                heading += (float)Posture.Legs[i].ShoulderYawAngle;

            Heading = heading / 6;
        }

        private static void FindRadius()
        {
            //Average the leg local (not rotated) X displacements to find the radius
            float radius = 0;
            for (int i = 0; i < 6; i++)
                radius += (float)Posture.Legs[i].X;

            Radius = radius / 6;
        }

        private static void FindStepHeight()
        {
            //find the difference in average depths between the two
            //leg groups.

            float Group0Depth = 0;
            float Group1Depth = 0;

            foreach (int LegIndex in LegGroup0)
                Group0Depth += Feet[LegIndex].Y;

            Group0Depth /= 3;

            foreach (int LegIndex in LegGroup1)
                Group1Depth += Feet[LegIndex].Y;

            Group1Depth /= 3;

            StepHeight = Group0Depth - Group1Depth;
        }

        private static void DoFootMap()
        {
            FootSensors[0] = Controller.LastCyclePacket.Feet[2];
            FootSensors[1] = Controller.LastCyclePacket.Feet[1];
            FootSensors[2] = Controller.LastCyclePacket.Feet[0];
            FootSensors[3] = Controller.LastCyclePacket.Feet[5];
            FootSensors[4] = Controller.LastCyclePacket.Feet[4];
            FootSensors[5] = Controller.LastCyclePacket.Feet[3];
        }
    }
}
