using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace ShoreSide.UI
{
    /// <summary>
    /// This class represents a histogram, not un-like the CPU graph in the Windows NT Task Manager.
    /// </summary>
    public partial class Histogram : Control
    {
        /// <summary>
        /// This is our "back buffer" which contains the current contents of the histogram.
        /// </summary>
        private Image backBuffer;

        /// <summary>
        /// The amount of time between updates.
        /// </summary>
        private TimeSpan updateDelay;

        /// <summary>
        /// This controls whether we draw the grid lines
        /// </summary>
        private bool drawGridLines;

        /// <summary>
        /// This controls the spacing between grid lines
        /// </summary>
        private int gridSpacing;

        /// <summary>
        /// This controls the colour of the grid.
        /// </summary>
        private Color gridColour;

        /// <summary>
        /// The grid offset ensures the background moves at the same rate as the
        /// rest of the histogram.
        /// </summary>
        private int gridOffset;

        /// <summary>
        /// This is the <see cref="Timer"/> object which controls updates to the histogram.
        /// </summary>
        private System.Windows.Forms.Timer updateTimer;

        /// <summary>
        /// The number of pixels the histogram moves to the left each update.
        /// </summary>
        private int moveSpeed;

        /// <summary>
        /// This is the list of histogram elements.
        /// </summary>
        private List<HistogramElement> histograms;

        /// <summary>
        /// This class contains the state for a single Histogram. The graph can display multiple
        /// histograms, so there could be many of these.
        /// </summary>
        private class HistogramElement
        {
            /// <summary>
            /// This is the amount of history we keep.
            /// </summary>
            private const int HISTORY_SIZE = 1000;

            /// <summary>
            /// This is the history for this histogram.
            /// </summary>
            private double[] history = new double[HISTORY_SIZE];

            /// <summary>
            /// This is our current position in the history. We use a "rolling array" so that we
            /// never have to allocate stuff.
            /// </summary>
            private int historyPosition;

            /// <summary>
            /// The colour of the histogram line...
            /// </summary>
            private Color colour;

            /// <summary>
            /// This is the delegate which samples the histogram when we go to update it.
            /// </summary>
            private SampleHistogram sample;

            /// <summary>
            /// This is the minimum value this histogram will be sampled at.
            /// </summary>
            private double minValue;

            /// <summary>
            /// This is the maximum value the histogram will be sampled at.
            /// </summary>
            private double maxValue;

            /// <summary>
            /// Constructs a new <see cref="HistoryElement"/>.
            /// </summary>
            public HistogramElement(SampleHistogram sample, Color colour, double minValue, double maxValue)
            {
                for (int i = 0; i < HISTORY_SIZE; i++)
                {
                    history[i] = Double.NaN;
                }

                // We start at one before the actual start of the histogram...
                historyPosition = HISTORY_SIZE - 1;

                this.sample = sample;
                this.colour = colour;
                this.minValue = minValue;
                this.maxValue = maxValue;
            }

            /// <summary>
            /// Gets the colour this histogram should be displayed as.
            /// </summary>
            public Color Colour
            {
                get { return colour; }
            }

            /// <summary>
            /// This is called to update this <see cref="HistogramElement"/>.
            /// </summary>
            public void Update()
            {
                historyPosition++;
                if (historyPosition >= HISTORY_SIZE)
                    historyPosition = 0;

                history[historyPosition] = sample();
            }

            /// <summary>
            /// Gets the value of the histogram at the given <c>position</c>.
            /// </summary>
            /// <param name="position">The position to get the value of. A <c>position</c> of 0 is
            /// the most recently-sampled data, 1 is the next most recent, etc.</param>
            /// <returns>The value of the histogram at the given position, normalized to a value
            /// between 0 and 1.</returns>
            public double GetValueAt(int position)
            {
                int index = historyPosition - position;
                while (index < 0)
                    index += HISTORY_SIZE;

                return (history[index] - minValue) / (maxValue - minValue);
            }
        }

        /// <summary>
        /// Constructs a new <see cref="Histogram"/>.
        /// </summary>
        public Histogram()
        {
            updateDelay = TimeSpan.FromSeconds(1.0);
            BackColor = Color.Black;
            drawGridLines = true;
            gridColour = Color.Green;
            gridSpacing = 10;
            moveSpeed = 2;
            updateTimer = new System.Windows.Forms.Timer();
            updateTimer.Interval = (int)updateDelay.TotalMilliseconds;
            updateTimer.Tick += new EventHandler(updateTimer_Tick);
            updateTimer.Enabled = false;
            if (!DesignMode)
            {
                // We only do updates if not in design mode...
                updateTimer.Enabled = true;
            }
            histograms = new List<HistogramElement>();

            InitializeComponent();
        }

        #region Public Properties
        /// <summary>
        /// Gets or sets the time between updates of the histogram.
        /// </summary>
        [Category("Behavior")]
        public TimeSpan UpdateDelay
        {
            get { return updateDelay; }
            set
            {
                updateDelay = value;
                if (!DesignMode)
                {
                    updateTimer.Interval = (int)updateDelay.TotalMilliseconds;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value that indicates whether we should draw the grid lines at all.
        /// </summary>
        [Category("Appearance")]
        public bool DrawGridLines
        {
            get { return drawGridLines; }
            set { drawGridLines = value; }
        }

        /// <summary>
        /// Gets or sets the number of pixels between the grid lines (if they're displayed at all).
        /// </summary>
        [Category("Appearance")]
        public int GridSpacing
        {
            get { return gridSpacing; }
            set
            {
                if (value <= 0)
                    throw new ArgumentException("GridSpacing must be >= 1", "value");

                gridSpacing = value;
            }
        }

        /// <summary>
        /// Gets or sets the colour of the grid lines (if they're displayed at all).
        /// </summary>
        [Category("Appearance")]
        public Color GridColour
        {
            get { return gridColour; }
            set { gridColour = value; }
        }

        /// <summary>
        /// Gets or sets the number of pixels the histogram moves with each update.
        /// </summary>
        [Category("Behavior")]
        public int MoveSpeed
        {
            get { return moveSpeed; }
            set { moveSpeed = value; }
        }

        /// <summary>
        /// We just override this because we don't want it to appear in the designer.
        /// </summary>
        [Browsable(false)]
        public override Font Font
        {
            get { return base.Font; }
            set { base.Font = value; }
        }

        /// <summary>
        /// We just override this because we don't want it to appear in the designer.
        /// </summary>
        [Browsable(false)]
        public override Color ForeColor
        {
            get { return base.ForeColor; }
            set { base.ForeColor = value; }
        }

        /// <summary>
        /// We just override this because we don't want it to appear in the designer.
        /// </summary>
        [Browsable(false)]
        public override string Text
        {
            get { return base.Text; }
            set { base.Text = value; }
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Adds a new histogram to the control.
        /// </summary>
        /// <param name="sample">The delegate which will be called to get the current value of the histogram.</param>
        /// <param name="colour">The colour of this histogram.</param>
        /// <param name="minValue">The minimum value the delegate will return.</param>
        /// <param name="maxValue">The maximum value the delegate will return.</param>
        public void AddHistogram(SampleHistogram sample, Color colour, double minValue, double maxValue)
        {
            histograms.Add(new HistogramElement(sample, colour, minValue, maxValue));
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// Paints the control. We compose the graph in an off-screen bitmap then copy the whole
        /// thing to the screen.
        /// </summary>
        /// <param name="pe">The arguments of the paint event.</param>
        protected override void OnPaint(PaintEventArgs pe)
        {
            if (backBuffer == null)
            {
                backBuffer = CreateBuffer();
            }

            pe.Graphics.DrawImageUnscaledAndClipped(backBuffer, ClientRectangle);
            base.OnPaint(pe);
        }

        /// <summary>
        /// Don't do anything here, because we do it all in <see cref="OnPaint"/>. This will avoid
        /// any flickering.
        /// </summary>
        /// <param name="pevent">The arguments for this event.</param>
        protected override void OnPaintBackground(PaintEventArgs pevent)
        {
        }

        /// <summary>
        /// This is called when the histogram is to be updated. We update all the samples, clear
        /// the back buffer and invalidate the rectangle.
        /// </summary>
        /// <param name="sender">The object which sent us this event (i.e. the timer object)</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>
        /// <para>The reason we simply clear the backbuffer and invalidte ourselves is because that
        /// way, if the control is hidden or minimized or something, then we won't actually do any
        /// drawing until we become visible again. Cool, eh?</para>
        /// </remarks>
        void updateTimer_Tick(object sender, EventArgs e)
        {
            if (DesignMode)
                return;

            gridOffset -= moveSpeed;
            while (gridOffset < 0)
                gridOffset += gridSpacing;

            foreach (HistogramElement histogram in histograms)
            {
                histogram.Update();
            }

            backBuffer = null;
            Invalidate();
        }

        /// <summary>
        /// This is called when the control is resized. We'll have to re-draw the backbuffer.
        /// </summary>
        /// <param name="e">The arguments for this event.</param>
        protected override void OnSizeChanged(EventArgs e)
        {
            // Just setting the back buffer to null causes it to be redrawn when we receive a
            // Paint event (which means we don't bother redrawing it if we're minimized or hidden,
            // etc).
            backBuffer = null;
            Invalidate();
        }

        /// <summary>
        /// This is called when the background colour changes. We'll have to redraw the background.
        /// </summary>
        /// <param name="e">The arguments for this event.</param>
        protected override void OnBackColorChanged(EventArgs e)
        {
            backBuffer = null;
            Invalidate();
        }
        #endregion

        #region Draw/Paint Methods
        /// <summary>
        /// Draws the Histogram onto a buffer, and returns the actual image.
        /// </summary>
        /// <returns>A new <see cref="Image"/> with the contents of the Histogram image.</returns>
        private Image CreateBuffer()
        {
            Image img = new Bitmap(Width, Height);
            Graphics g = Graphics.FromImage(img);
            Redraw(g, new Rectangle(0, 0, Width, Height));
            return img;
        }

        /// <summary>
        /// Draws the whole histogram with the given <see cref="Graphics"/> object.
        /// </summary>
        /// <param name="g">The <see cref="Graphics"/> object we use to draw the histogram.</param>
        private void Redraw(Graphics g, Rectangle size)
        {
            // First, draw a border around the histogram
            using (Pen pen = new Pen(Color.FromKnownColor(KnownColor.ControlLightLight)))
            {
                g.DrawRectangle(pen, size.X, size.Y, size.Width - 1, size.Height - 1);
            }
            using (Pen pen = new Pen(Color.FromKnownColor(KnownColor.ControlDark)))
            {
                g.DrawRectangle(pen, size.X, size.Y, size.Width - 2, size.Height - 2);
            }
            using (Pen pen = new Pen(Color.FromKnownColor(KnownColor.ControlLight)))
            {
                g.DrawRectangle(pen, size.X + 1, size.Y + 1, size.Width - 3, size.Height - 3);
            }
            using (Pen pen = new Pen(Color.FromKnownColor(KnownColor.ControlDarkDark)))
            {
                g.DrawRectangle(pen, size.X + 1, size.Y + 1, size.Width - 4, size.Height - 4);
            }

            // Now draw the background.
            using (Brush brush = new SolidBrush(BackColor))
            {
                g.FillRectangle(brush, size.X + 2, size.Y + 2, size.Width - 4, size.Height - 4);
            }

            // Draw the grid lines
            if (drawGridLines)
            {
                using (Pen pen = new Pen(gridColour))
                {
                    // Draw the vertical ones first
                    for (int x = size.Left + 2 + gridOffset; x < size.Right - 3; x += gridSpacing)
                    {
                        g.DrawLine(pen, x, 2, x, size.Height - 3);
                    }

                    // Now draw the horizontal ones
                    for (int y = size.Top + 2; y < size.Bottom - 3; y += gridSpacing)
                    {
                        g.DrawLine(pen, 2, y, size.Width - 3, y);
                    }
                }
            }

            // Finally, draw the actual histograms, cool!!
            foreach (HistogramElement histogram in histograms)
            {
                using (Pen pen = new Pen(histogram.Colour))
                {
                    int lastX = -1, lastY = -1;

                    int position = 0;
                    for (int x = size.Right - 3; x >= size.Left + 2; x -= moveSpeed)
                    {
                        double value = histogram.GetValueAt(position++);
                        if (Double.IsNaN(value))
                            break;

                        int y = size.Height - 3 - (int)(value * (size.Height - 6));

                        if (lastX != -1 && lastY != -1)
                            g.DrawLine(pen, x, y, lastX, lastY);

                        lastX = x;
                        lastY = y;
                    }
                }
            }
        }
        #endregion
    }

    /// <summary>
    /// This is the function which we call to sample the current histogram.
    /// </summary>
    /// <returns>The value to display in the histogram.</returns>
    public delegate double SampleHistogram();
}
