using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using ShoreSide.Core;

namespace ShoreSide
{
    public partial class ModelView : UserControl
    {
        private Device device;

        private static Vector3 CamPos = new Vector3(200, 200, 200);
        public static ModelBody body;

        public ModelView()
        {
            InitializeComponent();
        }

        protected void InitializeGraphics() 
		{ 
			this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
			
			PresentParameters pres = new PresentParameters(); 
			pres.Windowed = true; 
			pres.SwapEffect = SwapEffect.Discard; 
			pres.EnableAutoDepthStencil = true ; 
			pres.AutoDepthStencilFormat = DepthFormat.D16;

			int b,Q = 0;
			for ( int i = 0; i < (int) MultiSampleType.SixteenSamples; i ++)
			{
				if ( Manager.CheckDeviceMultiSampleType(0,DeviceType.Hardware,Manager.Adapters[0].CurrentDisplayMode.Format,true,(MultiSampleType) i,out b,out Q) )			
					pres.MultiSample = (MultiSampleType) i;			
			}
			pres.MultiSampleQuality = Q-1;

			device = new Device(0, DeviceType.Hardware, this , CreateFlags.HardwareVertexProcessing | CreateFlags.PureDevice , pres); 
			device.RenderState.CullMode = Cull.None; 
			device.Transform.Projection = Matrix.PerspectiveFovLH((float) Math.PI /2f,(float)this.Width/(float) this.Height ,1f, 200000f);

			// Texture filtering prevents textures to look bad when Magnified or Minified
			// Comment these lines out if your card does not support this function
			// To check wether your device supports Texture Filtering use the Manager class
			device.SamplerState[0].MagFilter = TextureFilter.Anisotropic;
			device.SamplerState[0].MinFilter = TextureFilter.Anisotropic;

			device.Transform.View = Matrix.LookAtLH(CamPos, new Vector3(0,0,0) ,new Vector3(0,1,0));	

			device.RenderState.Ambient = Color.Black;
			device.RenderState.CullMode = Cull.None; 

			device.Lights[0].Type = LightType.Point;
			device.Lights[0].Range = 1000;
			device.Lights[0].Attenuation0 = 0.9f;
			device.Lights[0].Attenuation1 = 0.0f;
			device.Lights[0].Position = new Vector3(150, 100, 150);
			device.Lights[0].Direction = new Vector3(-3,-3,-3);
			device.Lights[0].Ambient = Color.Gray;
			device.Lights[0].Specular = Color.White;
			device.Lights[0].Diffuse = Color.White;
			
			device.Lights[0].Update();
			
			device.Lights[0].Enabled = true;

		}

		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{	
			if (this.DesignMode)
				e.Graphics.FillRectangle(Brushes.Gray, 0, 0, this.Width, this.Height);
			else
			{
				Render();
			}				
		}	

		private void UpdateModel()
		{
			body.bearing = 0;
			body.position = new Vector3(0,0,0);
		}

		protected void Render() 
		{ 
			UpdateModel();

			// Don't render if the device isn't created or already disposed
			if ( device == null) return;
			if ( device.Disposed ) return;
			
			device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.Black , 1.0F, 0); 
			device.BeginScene();

			body.Render(Matrix.Identity);

			device.EndScene();
			device.Present();
		}

        private void ModelView_Load(object sender, EventArgs e)
        {
            if (this.DesignMode) return;

            InitializeGraphics();
            body = new ModelBody(device);
        }
    }

    public abstract class BodyPart
    {
        public BodyPart parent;
        public List<BodyPart> Children = new List<BodyPart>();
        public Matrix Transform = Matrix.Identity;
        public bool Enabled = true;

        protected Matrix MeshTransform = Matrix.Identity;
        protected Mesh PartMesh;
        protected Device device;
        protected Material mat = new Material();
        protected Matrix JointTransform = Matrix.Identity;

        public BodyPart(Device dev, BodyPart _parent)
        {
            device = dev;
            parent = _parent;

            mat.Ambient = Color.Black;
            mat.Diffuse = Color.Gray;
        }

        public void Render(Matrix ParentTransform)
        {
            device.Material = mat;

            this.UpdateJointTransform();
            Matrix tm = JointTransform * Transform * ParentTransform;
            device.Transform.World = MeshTransform * tm;

            if (Enabled)
                PartMesh.DrawSubset(0);

            foreach (BodyPart part in Children)
                part.Render(tm);
        }

        protected virtual void UpdateJointTransform()
        {
        }
    }

    public class ModelBody : BodyPart
    {
        public float bearing = 0.0F;
        public Vector3 position = new Vector3(0, 0, 0);
        public float pitch = 0.0F;
        public float roll = 0.0F;

        public LegSeg1[] legs = new LegSeg1[6];
        public LegHub[] hubs = new LegHub[6];
        public LegHub[] PlaneFeet = new LegHub[3];

        public GroundPlane ground = null;

        public ModelBody(Device dev)
            : base(dev, null)
        {
            this.PartMesh = Mesh.Cylinder(device, 110, 110, 40, 180, 10);
            this.MeshTransform = Matrix.RotationX((float)Math.PI / 2);

            this.mat.Diffuse = Color.Yellow;

            for (int LegNumber = 0; LegNumber < 6; LegNumber++)
            {
                legs[LegNumber] = new LegSeg1(device, this);
                legs[LegNumber].Transform = Matrix.Translation(135f, 0, 0) * Matrix.RotationY((float)LegNumber * (float)Math.PI / 3);
                this.Children.Add(legs[LegNumber]);

                hubs[LegNumber] = new LegHub(device, this);
                hubs[LegNumber].Transform = Matrix.Translation(120, 0, 0) * Matrix.RotationY((float)LegNumber * (float)Math.PI / 3);
                this.Children.Add(hubs[LegNumber]);

                if (LegNumber < 3)
                {
                    PlaneFeet[LegNumber] = new LegHub(device, this);
                    this.Children.Add(PlaneFeet[LegNumber]);
                }
            }

            ground = new GroundPlane(dev, this);
            this.Children.Add(ground);

            UpdateMatrix();
        }

        
        protected override void UpdateJointTransform()
        {
            Body b = Body.FromReadings();

            for (int i = 0; i < 6; i++)
            {
                Leg l = b.Legs[i];

                legs[i].Angle = (float)(l.ShoulderYawAngle * Math.PI / 180);
                legs[i].Seg2.Angle = (float)(l.ShoulderPitchAngle * Math.PI / 180);
                legs[i].Seg2.Seg3.Angle = (float)(l.ElbowAngle * Math.PI / 180);

                legs[i].CurrentSense = Servo.ServosByIndex[l.ShoulderYawIndex].CurrentReading;
                legs[i].Seg2.CurrentSense = Servo.ServosByIndex[l.ShoulderPitchIndex].CurrentReading;
                legs[i].Seg2.Seg3.CurrentSense = Servo.ServosByIndex[l.ElbowIndex].CurrentReading;
            }

            
            Vector3 Y = new Vector3(0, 1, 0);

            Plane plane = FindPlaneFromFeet(b);

            ground.Transform = Matrix.RotationX(plane.A) *
                Matrix.RotationY(plane.B) *
                Matrix.RotationZ(plane.C) * Matrix.Translation(0, -plane.D, 0);

        }

        private Plane FindPlaneFromFeet(Body b)
        {
            Plane p1 = Plane.FromPoints(VectorFromLeg(b.Legs[0]), VectorFromLeg(b.Legs[2]), VectorFromLeg(b.Legs[4]));
            Plane p2 = Plane.FromPoints(VectorFromLeg(b.Legs[1]), VectorFromLeg(b.Legs[3]), VectorFromLeg(b.Legs[5]));
            Plane c = new Plane((p1.A + p2.A) / 2, (p1.B + p2.B) / 2, (p1.C + p2.C) / 2, (p1.D + p2.D) / 2);
            return c;
        }

        private Vector3 VectorFromLeg(Leg leg)
        {
            Vector3 point = new Vector3((float)leg.X, (float)-leg.Y, (float)-leg.Z);
            point.TransformCoordinate(Matrix.Translation(135f, 0, 0) * Matrix.RotationY((float)leg.LegIndex * (float)Math.PI / 3));
            return point;
        }

        public void UpdateMatrix()
        {
            Transform = Matrix.RotationYawPitchRoll(this.bearing, this.pitch, this.roll) *
                Matrix.RotationY(this.bearing) *
                Matrix.Translation(this.position);
        }
    }

    public class GroundPlane : BodyPart
    {
        public GroundPlane(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.mat.Ambient = Color.Gray;
            this.mat.Diffuse = Color.Black;


            this.PartMesh = Mesh.Box(dev, 1000, 1, 1000);
        }
    }

    public class Foot : BodyPart
    {
        public Foot(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.mat.Ambient = Color.Black;
            this.mat.Diffuse = Color.White;
            this.mat.Emissive = Color.Gray;

            this.PartMesh = Mesh.Cylinder(dev, 9f, 6f, 24f, 10, 1);
            this.MeshTransform = Matrix.RotationX((float)Math.PI / 2);
        }
    }

    public class LegHub : BodyPart
    {
        public LegHub(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.mat.Ambient = Color.Black;
            this.mat.Diffuse = Color.Gray;

            this.PartMesh = Mesh.Cylinder(dev, 20, 20, 35f, 10, 1);
            this.MeshTransform = Matrix.RotationX((float)Math.PI / 2);
        }
    }

    public class LegSeg1 : BodyPart
    {
        public float Angle = 0.0f;
        public int CurrentSense = 0;

        public LegSeg2 Seg2 = null;

        public LegSeg1(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.PartMesh = Mesh.Box(device, 30, 30, 30);
            this.MeshTransform = Matrix.Translation(15f, 0, 0);

            Seg2 = new LegSeg2(device, this);
            Seg2.Transform = Matrix.Translation(30f, 0, 0);
            this.Children.Add(Seg2);
        }

        protected override void UpdateJointTransform()
        {
            this.JointTransform = Matrix.RotationY(-Angle);
            this.mat.Diffuse = Color.FromArgb(Math.Min(255, 128 + CurrentSense * 5), 128, 128);
        }
    }

    public class LegSeg2 : BodyPart
    {
        public float Angle = 0.0f;
        public int CurrentSense = 0;

        public LegSeg3 Seg3 = null;

        public LegSeg2(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.PartMesh = Mesh.Box(device, 70, 15f, 15f);
            this.MeshTransform = Matrix.Translation(35f, 0, 0);

            Seg3 = new LegSeg3(device, this);
            Seg3.Transform = Matrix.RotationZ((float)Math.PI / -2) * Matrix.Translation(70f, 0, 0);

            this.Children.Add(Seg3);
        }

        protected override void UpdateJointTransform()
        {
            this.JointTransform = Matrix.RotationZ(Angle);
            this.mat.Diffuse = Color.FromArgb(Math.Min(255, 128 + CurrentSense * 5), 128, 128);
        }
    }

    public class LegSeg3 : BodyPart
    {
        public float Angle = 0.0f;
        public int CurrentSense = 0;
        private Foot foot = null;

        public LegSeg3(Device dev, BodyPart parent)
            : base(dev, parent)
        {
            this.PartMesh = Mesh.Box(device, 110, 10, 10);
            this.MeshTransform = Matrix.Translation(55f, 0, 0);

            foot = new Foot(dev, this);
            foot.Transform = Matrix.RotationZ((float)Math.PI / 2) * Matrix.Translation(110, 0, 0);

            this.Children.Add(foot);

        }

        protected override void UpdateJointTransform()
        {
            this.JointTransform = Matrix.RotationZ(-Angle);
            this.mat.Diffuse = Color.FromArgb(Math.Min(255, 128 + CurrentSense * 5), 128, 128);
        }
    }
}
